/*
    md5mgr -- MD5 Checksum Manager

    Dieses Programm vereinfacht das Verwalten von MD5 Prfsummen in
    Verzeichnissbaumen.

    Copyright (C) Andreas Hofmeier

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License as
    published by the Free Software Foundation; either version 2 of
    the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
    General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.


    Name:     Andreas Hofmeier

    Adresse:  Stralsunder Strae 20
              28239 Bremen
              Germany

    Telefon:  +49-421-613793

    eMail:    An-H@An-H.de
              hofmeier@inbremen.de
              an-h@gmx.de

    Homepage: http://www.abmh.de/
	      hofmeier.home.pages.de
              http://www.LGut.uni-bremen.de/anh/

History
  v0.10: Caching hinzugefgt
  v0.12: 20.05.2008 Fixed BUG that program crashes when filename is longer that 60 bytes. In testfstats()

*/

#include <time.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <time.h>
#include "md5.h"


#define MAXLENFILENAME 2048
#define MAXLELDIRNAME 1024


// Logfiles
FILE *log_files,
  *log_md5sums,
  *log_files_new,
  *log_files_changed,
  *log_files_removed,
  *log_files_accerr;



char blacky_ver_info[] = "MD5 Checksum Manager v0.12 (c) Copyright 2003-2006 Andreas Hofmeier";

// Zhlervariablen fr Statistik
// ja, ich wei, globale Variablen sind kein guter Programmiertstiel ;-)
unsigned int s_dirs, s_files, s_ferr, s_fnew, s_fchanged, s_fmove, s_frem, s_filesi, s_fileso, s_mcha;


/* Erzeugt eine Fehlermeldung und schreibt diese in das LOG-file
   logfile: Logfile
   errormsg: Fehlermeldung
   dir: Verzeichnis in dem der Fehler aufgetreten ist
   file: Datei bei welcher der Fehler aufgetreten ist
   md5sum_file: Prfsumme von der Datei.
   md5sum_db: Prfsumme aus der Datenbank
*/
void emessage(FILE *logfile, char *errormsg, char *dir, char *file, char *md5sum_file, char *md5sum_db) {
  fprintf(logfile, "%s/%s: %s [%s -> %s]\n", dir, file, errormsg, md5sum_db, md5sum_file);
  fflush(logfile);
}


/* ffnet ein spetzielles Log-File zum Anhnhngen von Daten

   pname: Buffer fr Zugriff auf Main-DB und Logs
   pnamelen: lnge des Hauptdateinamens, welcher erweitert wird.
   logfile: Erweiterung des spetzellen logfiles
   mode: Aufrufsmodus
*/
FILE *log_init(char *pname, int pnamelen, char *logfile, int argc, char *argv[]) {
  FILE *f;
  unsigned long t;
  int i;

  strcpy(pname + pnamelen, logfile);
  if ((f = fopen(pname, "a")) == NULL) {
    perror(pname);
    exit(1);
  }

  time(&t);
  fprintf(f, "RUN: ");
  for (i = 0; i < argc; i++) {
    fprintf(f, "%s ", argv[i]);
  }
  fprintf(f, "\n AT: %s", ctime(&t));
  fflush(f);
  return f;
}


/* ffnet die Log-Files zum Anhnhngen von Daten

   pname: Buffer fr Zugriff auf Main-DB und Logs
   pnamelen: lnge des Hauptdateinamens, welcher erweitert wird.
   mode: Aufrufsmodus
*/
logs_init(char *pname, int pnamelen, int argc, char *argv[]) {
  log_files = log_init(pname, pnamelen, ".log.files", argc, argv);
  log_md5sums = log_init(pname, pnamelen, ".log.md5sums", argc, argv);
  log_files_new = log_init(pname, pnamelen, ".log.files_new", argc, argv);
  log_files_changed = log_init(pname, pnamelen, ".log.files_changed", argc, argv);
  log_files_removed = log_init(pname, pnamelen, ".log.files_removed", argc, argv);
  log_files_accerr = log_init(pname, pnamelen, ".log.files_accerr", argc, argv);
}

/* Schliet ein spetziellen logfile */
log_down(FILE *f) {
  unsigned long t;

  time(&t);
  fprintf(f, "DONE AT: %s", ctime(&t));
  fflush(f);
  fclose(f);
}

/* schliet alle logfiles */
logs_down() {
  log_down(log_files);
  log_down(log_md5sums);
  log_down(log_files_new);
  log_down(log_files_changed);
  log_down(log_files_removed);
  log_down(log_files_accerr);
}

// Entfernt die Newlines am Ende einer Zeile
void killenter(char *s) {
  int i;

  for (i = 0; ((s[i] != 0) && (s[i] != '\n') && (s[i] != 13)); i++) {
  }
  s[i] = 0;
}


/* Sucht einen Eintrag (fname) in einer md5db (dbname) und gibt die
   md5summe zurck
*/
char *getmd5sum(char *dbname, char *fname) {
  FILE *md5db;
  int BUFSIZE = MAXLENFILENAME + MAXLELDIRNAME + 2;
  static char buf[MAXLENFILENAME + MAXLELDIRNAME + 2];

  if ((md5db = fopen(dbname, "r")) == NULL) {
    buf[0] = 0;
    return buf;
  }

  while (!feof(md5db)) {
    buf[0] = 0;
    fgets(buf, BUFSIZE - 1, md5db);
    killenter(buf);

    if (strcmp(buf + 34, fname) == 0) {
      buf[32] = 0;

      fclose(md5db);
      return buf;
    }
  }
  fclose(md5db);
  buf[0] = 0;
  return buf;
}


/* Sucht eine md5summe (md5sum) in einer md5db (dbname) und gibt den
   dazugehrigen Dateinamen Zurck zurck
*/
char *getmd5sum_fname(char *dbname, char *md5sum) {
  FILE *md5db;
  int BUFSIZE = MAXLENFILENAME + MAXLELDIRNAME + 2;
  static char buf[MAXLENFILENAME + MAXLELDIRNAME + 2];

  if ((md5db = fopen(dbname, "r")) == NULL) {
    buf[0] = 0;
    return buf;
  }

  while (!feof(md5db)) {
    buf[0] = 0;
    fgets(buf, BUFSIZE - 1, md5db);
    killenter(buf);

    if (strncmp(buf, md5sum, 32) == 0) {
      fclose(md5db);
      return buf + 34;
    }
  }
  fclose(md5db);
  buf[0] = 0;
  return buf;
}


/* Fgt den Dateinamen (fname) mit der dazugehrigen Prfsumme
   (md5sum) an die MD5DB (dbname) an */
void addmd5sum(char *dbname, char *fname, char *md5sum) {
  FILE *f;

  if (md5sum[0] == 0) {
    return;
  }
  if ((f = fopen(dbname, "a")) == NULL) {
    perror(dbname);
    exit(1);
  }
  fprintf(f, "%s  %s\n", md5sum, fname);
  fflush(f);
  fclose(f);
}


/* ndert die zu einem Dateinamem (fname) dazugehrigen Prfsumme
   (md5sum) in der MD5DB (dbname) an */
changemd5sum(char *dbname, char *fname, char *md5sum, int caching) {
  FILE *s, *t;
  char dbname_tmp[strlen(dbname) + 2];
  int BUFSIZE = MAXLENFILENAME + MAXLELDIRNAME + 50;
  char buf[BUFSIZE];

  if (caching) {
    int namelen = strlen(dbname);
    strcat(dbname, ".cache");

    // open cache-file
    if ((t = fopen(dbname, "a")) == NULL) {
      perror(dbname);
      exit(1);
    }

    // neuen Eintrag anfgen -- cache
    if (md5sum[0] != 0) {
      fprintf(t, "%s  %s\n", md5sum, fname);
    } else {
      fprintf(t, "%s  %s\n", "  - to be removed -             ", fname);  
    }

    fflush(t);
    fclose(t);

    dbname[namelen] = 0;

    return;
  }

  // neue Datenbank, Eintrag anfgen
  if ((s = fopen(dbname, "r")) == NULL) {
    if ((t = fopen(dbname, "w")) == NULL) {
      perror(dbname);
      exit(1);
    }
    if (md5sum[0] != 0) {
      fprintf(t, "%s  %s\n", md5sum, fname);
    }
    fflush(t);
    fclose(t);
    return;
  } else {
    fclose(s);
  }

  strcpy(dbname_tmp, dbname);
  strcat(dbname_tmp, "~");

  if (rename(dbname, dbname_tmp) != 0) {
    perror("rename()");
    fprintf(stderr, "Can't move the file %s to %s\n", dbname, dbname_tmp);
    exit(1);
  }
  
  if ((s = fopen(dbname_tmp, "r")) == NULL) {
    perror(dbname_tmp);
    exit(1);
  }

  if ((t = fopen(dbname, "w")) == NULL) {
    perror(dbname);
    exit(1);
  }

  // alten Eintrag lschen
  while (!feof(s)) {
    buf[0] = 0;
    fgets(buf, BUFSIZE - 1, s);
    killenter(buf);
    if (buf[0] != 0) {
      if (strcmp(buf + 34, fname) != 0) {
	fputs(buf, t);
	fputs("\n", t);
      }
    }
  }    

  // neuen Eintrag anfgen
  if (md5sum[0] != 0) {
    fprintf(t, "%s  %s\n", md5sum, fname);
  }

  remove(dbname_tmp);
  fflush(t);
  fclose(t);
  fclose(s);
}


changemd5sum_transcache(char *dbname) {
  FILE *s, *t;
  char dbname_tmp[strlen(dbname) + 2];
  int BUFSIZE = MAXLENFILENAME + MAXLELDIRNAME + 50;
  char buf[BUFSIZE];
  int i;
  int entry_replaced;

  printf("changemd5sum_transcache(%s);\n", dbname);

  #define cache2size 25

  char cache[cache2size][BUFSIZE];

  strcpy(dbname_tmp, dbname);
  strcat(dbname_tmp, "~");

  int namelen = strlen(dbname);
  strcat(dbname, ".cache");

  FILE *fcache;

  // open cache-file
  if ((fcache = fopen(dbname, "r")) == NULL) {
    //    perror(dbname);
    //    exit(1);
    // nothing to do...
    return;
  }

  // db-file-name
  dbname[namelen] = 0;

  // neue Datenbank
  if ((s = fopen(dbname, "r")) == NULL) {
    if ((t = fopen(dbname, "w")) == NULL) {
      perror(dbname);
      exit(1);
    }

    // create empty file

    fflush(t);
    fclose(t);
  } else {
    fclose(s);
  }

  while (!feof(fcache)) {
    // fill cache 2 (memory) from cache 1 (file)
    for (i = 0; i < cache2size; i++) {
      cache[i][0] = 0;
      fgets(cache[i], BUFSIZE - 1, fcache);
      killenter(cache[i]);
    }

    if (rename(dbname, dbname_tmp) != 0) {
      perror("rename()");
      fprintf(stderr, "Can't move the file %s to %s\n", dbname, dbname_tmp);
      exit(1);
    }
  
    // open old-db file
    if ((s = fopen(dbname_tmp, "r")) == NULL) {
      perror(dbname_tmp);
      exit(1);
    }

    // open new-db-file
    if ((t = fopen(dbname, "w")) == NULL) {
      perror(dbname);
      exit(1);
    }

    // alten Eintrag lschen
    while (!feof(s)) {
      buf[0] = 0;
      fgets(buf, BUFSIZE - 1, s);
      killenter(buf);
      if (buf[0] != 0) {
	entry_replaced = 0;

	for (i = 0; i < cache2size; i++) {
	  // replace entry, when filename is equal
	  if (cache[i][0] != 0) {
	    if (strcmp(buf + 34, cache[i] + 34) == 0) {
	      entry_replaced = 1;
	      break;
	    }
	  }
	}

	if (entry_replaced == 0) {
	  fputs(buf, t);
	  fputs("\n", t);
	}
      }
    }
    
    // neuen Eintrag anfgen
    for (i = 0; i < cache2size; i++) {
      if ((cache[i][0] != 0) && (cache[i][0] != ' ')) {
	fputs(cache[i], t);
	fputs("\n", t);
      }
    }
    
    remove(dbname_tmp);
    fflush(t);
    fclose(t);
    fclose(s);
  }
  
  fclose(fcache);

  strcat(dbname, ".cache");
  remove(dbname);
}


/* Eintfernt alle Datei-Eintrge fr ein Verzeichnis (dirname) aus einer
   Haupt MD5Sum-DB (pname . dbname).
*/
void removing_all_entrys_of_a_directory_from_main_db(char *pname, int pnamelen, char *dbname, char *dirname) {
  FILE *s, *t;
  char dbname_tmp[strlen(dbname) + 2];
  int BUFSIZE = MAXLENFILENAME + MAXLELDIRNAME + 50;
  char buf[BUFSIZE];
  int dirnamelen = strlen(dirname);

  strcpy(dbname_tmp, dbname);
  strcat(dbname_tmp, "~");

  if ((s = fopen(dbname, "r")) == NULL) {
    return;
  } else {
    fclose(s);
  }

  // aktuelle DB verschieben.
  if (rename(dbname, dbname_tmp) != 0) {
    perror("rename();");
    fprintf(stderr, "Can't move the file %s to %s\n", dbname, dbname_tmp);
    exit(1);
  }
  
  // alte DB ffnen, neue erzeugen
  if ((s = fopen(dbname_tmp, "r")) == NULL) {
    perror(dbname_tmp);
    exit(1);
  }

  if ((t = fopen(dbname, "w")) == NULL) {
    perror(dbname);
    exit(1);
  }

  // Inhalte von alt in neu copieren und dabei Eintrge lschen
  while (!feof(s)) {
    buf[0] = 0;
    fgets(buf, BUFSIZE - 1, s);
    killenter(buf);
    if (buf[0] != 0) {
      if (!((strncmp(buf + 34, dirname, dirnamelen) == 0) &&
	    (strstr(buf + 34 + dirnamelen + 1, "/") == NULL))) {
	fputs(buf, t);
	fputs("\n", t);
      } else {
	s_frem++;
	buf[32] = 0;
	emessage(log_files, "file removed", "", buf + 34, "", buf);
	fputs(buf + 34, log_files_removed);
	fputs("\n", log_files_removed);
	fflush(log_files_removed);
      }
    }
  }

  remove(dbname_tmp);
  fflush(t);
  fclose(t);
  fclose(s);
}


/* Versucht aus der Haupt MD5-Datenbank (pname . .md5sums) eine lokale
   (dirname . "/.md5sums") zu rekonstruieren. Dazu werden alle Eintrage
   von Dateien aus einem Verzeichnis (dirname) ausgelesen, in relative
   umgeformt und in die locale DB geschrieben.
*/
void recover_local_md5sum(char *pname, int pnamelen, char *dirname, int dirnamelen) {
  FILE *s, *t;
  int BUFSIZE = MAXLENFILENAME + MAXLELDIRNAME + 50;
  char buf[BUFSIZE];

  strcpy(pname + pnamelen, ".md5sums");

  if ((s = fopen(pname, "r")) == NULL) {
    perror(pname);
    //    exit(1);
    return;
  }

  strcpy(dirname + dirnamelen, "/.md5sums");
  if ((t = fopen(dirname, "w")) == NULL) {
    perror(dirname);
    exit(1);
  }
  dirname[dirnamelen] = 0;

  while (!feof(s)) {
    buf[0] = 0;
    fgets(buf, BUFSIZE - 1, s);
    killenter(buf);
    if ((buf[0] != 0) &&
	(strncmp(buf + 34, dirname, dirnamelen) == 0) &&
	(strstr(buf + 34 + dirnamelen + 1, "/") == NULL)) {
      buf[32] = 0;
      fprintf(t, "%s  %s\n", buf, buf + 34 + dirnamelen + 1);
    }
  }
  fflush(t);
  fclose(t);
  fclose(s);
}


// Erzeugt eine MD5summe von einer Datei (fname) und gibt diese zurck
char *makemd5sum(char *fname) {
  FILE *f;
  char command[MAXLENFILENAME + MAXLELDIRNAME + 50];
  static char buf[33];

  int hrbufsize = 2048;
  int hread;
  char hrbuf[hrbufsize];

  struct MD5Context context;
  unsigned char md5_prs[16];
  int i;

  if ((f = fopen(fname, "r")) == NULL) {
    buf[0] = 0;
    return buf;
  }

  MD5Init(&context);
  while (!feof(f)) {
    hread = fread(hrbuf, 1, hrbufsize, f);
    MD5Update(&context, hrbuf, hread);
  }
  fclose(f);
  MD5Final(md5_prs, &context);

  for (i = 0; i < 16; i++) {
    sprintf(buf + (i*2), "%02x", md5_prs[i]);
  }

  buf[32] = 0;
  return buf;
}


/* berptft, ob die Angaben fr eine Datei (filename) und deren
   Prfsumme (md5sum) im fstats-file (dirname . /.md5mgr.fstats)
   gleich denen aktuellen sind
*/
int testfstats(char *pname, int pnamelen, char *dirname, char *filename, int mode, char *md5sum, int simple, int relax) {
  struct stat statf;
  FILE *f;
  char mbuf[33 + 60];
  char b0[80], b1[80];
  int changed = 0;
  int changed_relax = 0;
  int blocksize = sizeof(mode_t) + sizeof(nlink_t) + sizeof(uid_t) + sizeof(gid_t) + sizeof(off_t) + 2 * sizeof(time_t);
  mode_t        t_mode;     /* protection */
  nlink_t       t_nlink;    /* number of hard links */
  uid_t         t_uid;      /* user ID of owner */
  gid_t         t_gid;      /* group ID of owner */
  off_t         t_size;     /* total size, in bytes */
  time_t        t_mtime;    /* time of last modification */
  time_t        t_ctime;    /* time of last change */
  char fnamebuf[60];

  // im einfachen Modus nur berprfen, ob die Datei vorhenden ist
  if (simple) {
    if ((f = fopen(filename, "r")) == NULL) {
      return 1;
    } else {
      fclose(f);
      return 0;
    }
  }

  // Dateinamen in 60-Byte-Array einpassen
  //  char fnamebuf[60];
  {
    int i, len = strlen(filename);
    // only the first 60-bytes of a filenameare stored in the stats-file.
    if (len >= 60) {
      len = 60;
    }
    
    for (i = 0; i < len; i++) {
      fnamebuf[i] = filename[i];
    }
    for (  ; i < 60; i++) {
      fnamebuf[i] = 0;
    }
  }

  // aktuellen Status laden
  if (stat(filename, &statf) != 0) {
    perror(filename);
    return 1;
  }

  // falls keine DB vorhanden anlegen und anfgen
  if (mode == 2) {
    if ((f = fopen("./.md5mgr.fstats", "r")) == NULL) {
      return 1;
    }
  } else {
    if ((f = fopen("./.md5mgr.fstats", "r+")) == NULL) {
      if ((f = fopen("./.md5mgr.fstats", "w")) == NULL) {
	return 1;
      } else {
	// an db anfgen
	emessage(log_files, "fstat difference: file not in db", dirname, filename, "", "");
	fwrite(md5sum, 32, 1, f);
	fwrite(fnamebuf, 60, 1, f);
	fwrite(&statf.st_mode, sizeof(statf.st_mode), 1, f);
	fwrite(&statf.st_nlink, sizeof(statf.st_nlink), 1, f);
	fwrite(&statf.st_uid, sizeof(statf.st_uid), 1, f);
	fwrite(&statf.st_gid, sizeof(statf.st_gid), 1, f);
	fwrite(&statf.st_size, sizeof(statf.st_size), 1, f);
	fwrite(&statf.st_mtime, sizeof(statf.st_mtime), 1, f);
	fwrite(&statf.st_ctime, sizeof(statf.st_ctime), 1, f);
	fflush(f);
	fclose(f);
	return 1;
      }
    }
  }

  // Daten in Datenbank suchen, vergleichen und anpassen
  while (!feof(f)) {
    mbuf[0] = 0;
    fread(mbuf, 32 + 60, 1, f);
    if (mbuf[0] == 0) {
      break;
    }
    mbuf[32 + 60] = 0;
    if ((strncmp(mbuf, md5sum, 32) == 0) && (strncmp(mbuf + 32, fnamebuf, 60) == 0)) {
      fread(&t_mode, sizeof(t_mode), 1, f);
      fread(&t_nlink, sizeof(t_nlink), 1, f);
      fread(&t_uid, sizeof(t_uid), 1, f);
      fread(&t_gid, sizeof(t_gid), 1, f);
      fread(&t_size, sizeof(t_size), 1, f);
      fread(&t_mtime, sizeof(t_mtime), 1, f);
      fread(&t_ctime, sizeof(t_ctime), 1, f);

      if (t_mode != statf.st_mode) {
	sprintf(b0, "%d", t_mode);
	sprintf(b1, "%d", statf.st_mode);
	emessage(log_files, "fstat difference: protection", dirname, filename, b1, b0);
	changed_relax = 1;
      }
      if (t_nlink != statf.st_nlink) {
	sprintf(b0, "%d", t_nlink);
	sprintf(b1, "%d", statf.st_nlink);
	emessage(log_files, "fstat difference: number of hard links", dirname, filename, b1, b0);
	changed_relax = 1;
      }
      if (t_uid != statf.st_uid) {
	sprintf(b0, "%d", t_uid);
	sprintf(b1, "%d", statf.st_uid);
	emessage(log_files, "fstat difference: user ID of owner", dirname, filename, b1, b0);
	changed_relax = 1;
      }
      if (t_gid != statf.st_gid) {
	sprintf(b0, "%d", t_gid);
	sprintf(b1, "%d", statf.st_gid);
	emessage(log_files, "fstat difference: group ID of owner", dirname, filename, b1, b0);
	changed_relax = 1;
      }
      if (t_size != statf.st_size) {
	sprintf(b0, "%d", t_size);
	sprintf(b1, "%d", statf.st_size);
	emessage(log_files, "fstat difference: total size, in bytes", dirname, filename, b1, b0);
	changed = 1;
      }
      // nderung des Dateiinhaltes
      if (t_mtime != statf.st_mtime) {
	strcpy(b0, ctime(&(t_mtime)));
	killenter(b0);
	strcpy(b1, ctime(&(statf.st_mtime)));
	killenter(b1);
	emessage(log_files, "fstat difference: time of last modification", dirname, filename, b1, b0);
	//	changed = 1;
	changed_relax = 1;
      }
      // nderung der Datei im System, z.B. nach cp -a ...
      if (t_ctime != statf.st_ctime) {
	strcpy(b0, ctime(&(t_ctime)));
	killenter(b0);
	strcpy(b1, ctime(&(statf.st_ctime)));
	killenter(b1);
	emessage(log_files, "fstat difference: time of last change", dirname, filename, b1, b0);
	changed_relax = 1;
      }

      if ((changed) || (changed_relax)) {
	// db ndern
	if (mode != 2) {
	  fseek(f, (long) (-1 * blocksize), SEEK_CUR);
	  fwrite(&statf.st_mode, sizeof(statf.st_mode), 1, f);
	  fwrite(&statf.st_nlink, sizeof(statf.st_nlink), 1, f);
	  fwrite(&statf.st_uid, sizeof(statf.st_uid), 1, f);
	  fwrite(&statf.st_gid, sizeof(statf.st_gid), 1, f);
	  fwrite(&statf.st_size, sizeof(statf.st_size), 1, f);
	  fwrite(&statf.st_mtime, sizeof(statf.st_mtime), 1, f);
	  fwrite(&statf.st_ctime, sizeof(statf.st_ctime), 1, f);
	}
	fflush(f);
	fclose(f);
	if ((changed) || ((changed_relax) && (!relax))) {
	  return 1;
	} else {
	  return 0;
	}
      } else {
	fclose(f);
	return 0;
      }
    }
    if (fseek(f, (long) (blocksize), SEEK_CUR) != 0) {
      break;
    }
  }
  // an db anfgen
  emessage(log_files, "fstat difference: file not in db", dirname, filename, "", "");
  if (mode != 2) {
    fwrite(md5sum, 32, 1, f);
    fwrite(fnamebuf, 60, 1, f);
    fwrite(&statf.st_mode, sizeof(statf.st_mode), 1, f);
    fwrite(&statf.st_nlink, sizeof(statf.st_nlink), 1, f);
    fwrite(&statf.st_uid, sizeof(statf.st_uid), 1, f);
    fwrite(&statf.st_gid, sizeof(statf.st_gid), 1, f);
    fwrite(&statf.st_size, sizeof(statf.st_size), 1, f);
    fwrite(&statf.st_mtime, sizeof(statf.st_mtime), 1, f);
    fwrite(&statf.st_ctime, sizeof(statf.st_ctime), 1, f);
    fflush(f);
  }
  fclose(f);
  return 1;
}


/* Entfernt einen Eintrag (filename und md5sum) aus der localen
   ./.md5mgr.fstats */
removefstats(char *md5sum, char *filename) {
  FILE *s, *t;
  //  char mbuf[33];
  int blocksize = sizeof(mode_t) + sizeof(nlink_t) + sizeof(uid_t) + sizeof(gid_t) + sizeof(off_t) + 2 * sizeof(time_t);
  char buf[blocksize + 33 + 60];

  char nbuf[blocksize + 33 + 60];
  char *fname;

  int i, len;

  if ((s = fopen("./.md5mgr.fstats", "r")) == NULL) {
    return;
  } else {
    fclose(s);
  }

  // aktuelle DB verschieben
  if (rename("./.md5mgr.fstats", "./.md5mgr.fstats.tmp") != 0) {
    perror("rename();");
    fprintf(stderr, "Can't move the file ./.md5mgr.fstats to ./.md5mgr.fstats.tmp\n");
    exit(1);
  }

  // alte ffnen, neue anlegen
  if ((s = fopen("./.md5mgr.fstats.tmp", "r")) == NULL) {
    perror("./.md5mgr.fstats.tmp");
    exit(1);
  }

  if ((t = fopen("./.md5mgr.fstats", "w")) == NULL) {
    perror("./.md5mgr.fstats");
    exit(1);
  }

  // alle Eintrge kopieren, auer den zu entfernenden natrlich
  while (!feof(s)) {
    buf[0] = 0;
    fread(buf, blocksize + 32 + 60, 1, s);
    if (buf[0] != 0) {
      len = strlen(filename);
      if (len > 60) {
	len = 60;
      }
      if (!(
	    (strncmp(buf, md5sum, 32) == 0) &&
	    (strncmp(buf + 32, filename, len) == 0))) {
	fwrite(buf, blocksize + 32 + 60, 1, t);
      }
    }
  }
 
  /*
  while (!feof(s)) {
    buf[0] = 0;
    fread(buf, blocksize + 32, 1, s);
    if (buf[0] != 0) {
      strncpy(nbuf, buf, 32);
      nbuf[32] = 0;
      fname = getmd5sum_fname("./.md5sums", nbuf);
      
      len = strlen(fname);
      for (i = 0; i < len; i++) {
	nbuf[i + 32] = fname[i];
      }
      for (  ; i < 60; i++) {
	nbuf[i + 32] = 0;
      }

      for (i = 32 + 60; i < blocksize + 32 + 60; i++) {
	nbuf[i] = buf[i - (60)];
      }
      fwrite(nbuf, blocksize + 32 + 60, 1, t);
    }
  }
  */

  remove("./.md5mgr.fstats.tmp");
  fflush(t);
  fclose(t);
  fclose(s);
}


/* Verarbeite ein Verzeichnis (dirname) */
void dodir(char *pname, int pnamelen, char *dirname, int mode, int local, int simple, int trustnew, int relax) {
  FILE *md5sum_dir, *p;
  char md5sum_file[33], md5sum_db[33];
  unsigned char localMD5DBchanged = 0;
  unsigned char localMD5FSTATSchanged = 0;

  int BUFSIZE = MAXLENFILENAME + MAXLELDIRNAME + 50;
  char buf[BUFSIZE];
  int dirnamelen;

  printf("examining: %s\n", dirname);
  fflush(stdout);

  // Slash am ande Entfernen, falls vorhanden
  dirnamelen = strlen(dirname);
  if (dirname[dirnamelen - 1] == '/') {
    dirname[--dirnamelen] = 0;
  }
  //  int pnamelen = strlen(pname);

  dirname[dirnamelen] = 0;
  if (chdir(dirname) != 0) {
    perror(dirname);
    emessage(log_files, "Can't change directory", dirname, "", "", "");
    return;
  }

  s_dirs++;

  if (!local) {
    // berprfe md5sums-file an sich mit der Hauptdatenbank .2md5sums
    strcpy(dirname + dirnamelen, "/.md5sums");
    strcpy(pname + pnamelen, ".2md5sums");
    strcpy(md5sum_db, getmd5sum(pname, dirname));
    strcpy(md5sum_file, makemd5sum(dirname));
    // locale md5sum nicht vorhanden ader in DB gemeldet, versuche Wiederherstellung
    if ((md5sum_dir = fopen(dirname, "r")) == NULL) {
      if (md5sum_db[0] != 0) {
	dirname[dirnamelen] = 0;
	puts("md5sum-file removed trying recover..\n");
	emessage(log_md5sums, "md5sum-file removed trying recover..", dirname, "", md5sum_file, md5sum_db);
	recover_local_md5sum(pname, pnamelen, dirname, dirnamelen);
	s_mcha++;
      }
    } else {
      // locale md5sum vorhanden ader nicht in DB gemeldet, vertraue
      // neuer Datei
      if ((md5sum_db[0] == 0) && (trustnew)) {
	if ((mode != 2) && (local == 0)) {
	  emessage(log_md5sums, "trust new md5sum-file", dirname, "", md5sum_file, md5sum_db);
	  printf("trust new md5sums.\n");
	
	  //	  strcpy(dirname + dirnamelen, buf + 1);
	  strcpy(pname + pnamelen, ".md5sums");

	  localMD5DBchanged = 1;

	  // fge Prfsummen auf neuer DB an Haupt DB an
	  while (!feof(md5sum_dir)) {
	    buf[0] = 0;
	    fgets(buf, BUFSIZE - 1, md5sum_dir);
	    if (buf[0] != 0) {
	      killenter(buf);
	      buf[32] = 0;
	      strcpy(dirname + dirnamelen, "/");
	      strcpy(dirname + dirnamelen + 1, buf + 34);
	      if (mode != 2) {
		addmd5sum(pname, dirname, buf);
	      }
	    }
	  }
	}
	fclose(md5sum_dir);
      } else {
	fclose(md5sum_dir);

	// Prfsumme der md5sums-Datei ist nicht korrekt,
	// wiederherstellen...
	if (strcmp(md5sum_db, md5sum_file) != 0) {
	  //	dirname[dirnamelen] = 0;
	  strcpy(dirname + dirnamelen, "/.md5sums");
	  emessage(log_md5sums, "md5sum-file changed trying recover", dirname, "", md5sum_file, md5sum_db);
	  //	remove(dirname);
	  printf("md5sums changed or new. md5sums moved to md5sums_org. trying recover from maindb...\n");
	  if (rename(".md5sums", ".md5sums_org") != 0) {
	    perror("rename();");
	    exit(1);
	  }
	  recover_local_md5sum(pname, pnamelen, dirname, dirnamelen);
	  s_mcha++;
	}
      }
    }

    // lche .md5mgr.fstats, falls .md5sums nicht vorhanden ist und
    // auch nicht wieder hergestellt werden kann
    if ((md5sum_db[0] == 0) && (md5sum_file[0] == 0)) {
      remove(".md5mgr.fstats");
    }
    
    if (!simple) {
      // berprfe .md5mgr.fstats
      strcpy(dirname + dirnamelen, "/.md5mgr.fstats");
      strcpy(pname + pnamelen, ".2md5mgr.fstats");
      strcpy(md5sum_db, getmd5sum(pname, dirname));
      strcpy(md5sum_file, makemd5sum(dirname));
      if (strcmp(md5sum_db, md5sum_file) != 0) {
	//	strcpy(dirname + dirnamelen, "/.md5sums");
	if ((md5sum_db[0] == 0) && (md5sum_file[0] != 0) && (trustnew)) {
	  emessage(log_md5sums, "trust new .md5mgr.fstats!", dirname, "", md5sum_file, md5sum_db);
	  printf("trust new md5mgr.fstats file\n");
	  localMD5FSTATSchanged = 1;
	} else {
	  emessage(log_md5sums, ".md5mgr.fstats changed!", dirname, "", md5sum_file, md5sum_db);
	  remove(dirname);
	  printf("md5mgr.fstats changed\n");
	}
      }
    }
  }



  // sichere Zugriffszeit fr sptere Verarbeitung
  system("touch -r .md5sums .md5sums.timedate 2> /dev/null");
  // gehe alle Files im aktuellem Verzeichnis durch und berprfe
  // diese
  if ((p = popen("find . -type f -maxdepth 1 -not -name \\.md5sums\\* -not -name .md5mgr.fstats", "r")) == NULL) {
    //    dirname[dirnamelen] = 0;
    emessage(log_files, "Can't get a filelist", dirname, "", "", "");
  } else {
    while (!feof(p)) {
      buf[0] = 0;
      fgets(buf, BUFSIZE - 1, p);
      if (buf[0] != 0) {
	s_files++;
	killenter(buf);
	// try get md5sum from local db
	strcpy(md5sum_db, getmd5sum(".md5sums", buf + 2));
	
	if (md5sum_db[0] == 0) {
	  
	  if (0) {
	  //	  if (local == 0) {
	    // try get md5sum from main db
	    strcpy(dirname + dirnamelen, buf + 1);
	    strcpy(pname + pnamelen, ".md5sums");
	    strcpy(md5sum_db, getmd5sum(pname, dirname));
	    if (mode != 2) {
	      localMD5DBchanged = 1;
	      addmd5sum(".md5sums", buf + 2, md5sum_db);
	    }
	  }
	}
	if (md5sum_db[0] == 0) {
	  // md5summe von dem File erzeugen
	  strcpy(md5sum_file, makemd5sum(buf));
	  if (md5sum_file[0] == 0) {
	    dirname[dirnamelen] = 0;
	    emessage(log_files, "Can't make md5sum from", dirname, buf + 2, "", "");
	    fprintf(log_files_accerr, "%s/%s\n", dirname, buf + 2);
	    fflush(log_files_accerr);

	    printf("X");
	    fflush(stdout);
	    s_ferr++;
	  } else {
	    char *filename = NULL;

	    if (!simple) {
	      // fstats aufnehmen
	      dirname[dirnamelen] = 0;
	      testfstats(pname, pnamelen, dirname, buf + 2, mode, md5sum_file, simple, relax);
	      localMD5FSTATSchanged = 1;
	    }
	    
	    // berprfe, ob eine Datei nur Verschoben wurde
	    strcpy(pname + pnamelen, ".md5sums");
	    // *** zu langsam ***
	    //	    filename[0] = "";
	    //	    filename = getmd5sum_fname(pname, md5sum_file);
	    if ((filename != NULL) && (filename[0] != 0)) {
	      strcpy(dirname + dirnamelen, buf + 1);
	      emessage(log_files, "file moved or have the same md5sum", md5sum_file, "", dirname, filename);
	      //	      fprintf(log_files_changed, "%s/%s\n", dirname, filename);
	      //	      fflush(log_files_changed);
	      printf("m");
	      fflush(stdout);
	      s_fmove++;
	    } else {		
	      dirname[dirnamelen] = 0;
	      emessage(log_files, "new file", dirname, buf + 2, md5sum_file, "");
	      fprintf(log_files_new, "%s/%s\n", dirname, buf + 2);
	      fflush(log_files_new);
	      printf("n");
	      fflush(stdout);
	      s_fnew++;
	    }
	    if (mode != 2) {
	      localMD5DBchanged = 1;
	      // an locale db anfgen, an globale db anfgen
	      addmd5sum(".md5sums", buf + 2, md5sum_file);
	      if (local == 0) {
		strcpy(dirname + dirnamelen, buf + 1);
		strcpy(pname + pnamelen, ".md5sums");
		addmd5sum(pname, dirname, md5sum_file);
	      }
	    }
	  }
	} else {
	  dirname[dirnamelen] = 0;
	  if (mode != 1 | (testfstats(pname, pnamelen, dirname, buf + 2, mode, md5sum_db, simple, relax) == 1) ? localMD5FSTATSchanged = 1 : 0) {
	    strcpy(md5sum_file, makemd5sum(buf));
	    // konnte Prfsumme von Datei nicht erzeugen
	    if (md5sum_file[0] == 0) {
	      dirname[dirnamelen] = 0;
	      emessage(log_files, "Can't make md5sum from", dirname, buf + 2, "", "");
	      fprintf(log_files_accerr, "%s/%s\n", dirname, buf + 2);
	      fflush(log_files_accerr);
	      printf("X");
	      fflush(stdout);
	      s_ferr++;
	    } else {
	      // Datei wurde gendert 
	      if (strcmp(md5sum_file, md5sum_db) != 0) {

		if (!simple) {
		  if (mode != 2) {
		    removefstats(md5sum_db, buf + 2);
		  }
		  dirname[dirnamelen] = 0;

		  testfstats(pname, pnamelen, dirname, buf + 2, mode, md5sum_file, simple, relax);
		  localMD5FSTATSchanged = 1;
		}

		dirname[dirnamelen] = 0;
		emessage(log_files, "file changed", dirname, buf + 2, md5sum_db, md5sum_file);
		fprintf(log_files_changed, "%s/%s\n", dirname, buf + 2);
		fflush(log_files_changed);
		printf("c");
		fflush(stdout);
		s_fchanged++;
		if (mode != 2) {
		  localMD5DBchanged = 1;
		  // in localer und globeler db ndern
		  changemd5sum(".md5sums", buf + 2, md5sum_file, 0);
		  if (local == 0) {
		    strcpy(dirname + dirnamelen, buf + 1);
		    strcpy(pname + pnamelen, ".md5sums");
		    changemd5sum(pname, dirname, md5sum_file, 1);
		  }
		}
	      } else {
		s_fileso++;
		printf(".");
		fflush(stdout);
	      }
	    }
	  } else {
	    s_filesi++;
	    printf("-");
	    fflush(stdout);
	  }
	}
      }
    }
    pclose(p);
  
    printf("\n");
    fflush(stdout);

  reopen_db:
    // berprfen, ob alle files in der DB auch im Verzeichnis sind
    if ((md5sum_dir = fopen(".md5sums", "r")) == NULL) {
      //      perror("md5sum");
      //      exit(1);
    } else {
      while (!feof(md5sum_dir)) {
	buf[0] = 0;
	fgets(buf, BUFSIZE - 1, md5sum_dir);
	killenter(buf);
	if (buf[0] == 0) {
	  break;
	}
	
	if ((p = fopen(buf + 34, "r")) == NULL) {
	  dirname[dirnamelen] = 0;
	  strcpy(md5sum_db, getmd5sum(".md5sums", buf + 34));
	  if (!simple) {
	    if (mode != 2) {
	      removefstats(md5sum_db, buf + 34);
	    }
	    localMD5FSTATSchanged = 1;
	  }
	  emessage(log_files, "file removed", dirname, buf + 34, "", md5sum_db);
	  fprintf(log_files_removed, "%s/%s\n", dirname, buf + 34);
	  fflush(log_files_removed);
	  printf("r");
	  fflush(stdout);
	  s_frem++;
	  if (mode != 2) {
	    localMD5DBchanged = 1;
	    // in localer und globeler db lschen
	    changemd5sum(".md5sums", buf + 34, "", 0);
	    if (local == 0) {
	      strcpy(dirname + dirnamelen, "/");
	      strcpy(dirname + dirnamelen + 1, buf + 34);
	      strcpy(pname + pnamelen, ".md5sums");
	      changemd5sum(pname, dirname, "", 1);
	    }
	    
	    fclose(md5sum_dir);
	    goto reopen_db;
	  }
	} else {
	  fclose(p);
	}
      }
      fclose(md5sum_dir);
    }

    printf("\n");
    fflush(stdout);

    // normalerweise werden genderte files via .md5mgr.fstats
    // erkannt. In diesem Modi nicht, daher:
    if (simple) {
      // suche auf grund der Systemzeit nach genderten Files -- bei
      // fullmode (testmode is fullmode) natrlich berflssig
      if (mode == 1) {
	
	if ((p = popen("find . -type f -maxdepth 1 -newer .md5sums.timedate -not -name \\.md5sums\\* -not -name .md5mgr.fstats 2> /dev/null", "r")) == NULL) {
	  //    dirname[dirnamelen] = 0;
	  emessage(log_files, "Can't get a filelist -- -newer", dirname, "", "", "");
	} else {
	  while (!feof(p)) {
	    buf[0] = 0;
	    fgets(buf, BUFSIZE - 1, p);
	    if (buf[0] != 0) {
	      killenter(buf);
	      // try get md5sum from local db
	      strcpy(md5sum_db, getmd5sum(".md5sums", buf + 2));
	      if (md5sum_db[0] == 0) {
		if (local == 0) {
		  // try get md5sum from main db
		  strcpy(dirname + dirnamelen, buf + 1);
		  strcpy(pname + pnamelen, ".md5sums");
		  strcpy(md5sum_db, getmd5sum(pname, dirname));
		  if (mode != 2) {
		    localMD5DBchanged = 1;
		    addmd5sum(".md5sums", buf + 2, md5sum_db);
		  }
		}
	      }
	      if (md5sum_db[0] == 0) {
		// md5summe von dem File erzeugen
		strcpy(md5sum_file, makemd5sum(buf));
		if (md5sum_file[0] == 0) {
		  dirname[dirnamelen] = 0;
		  emessage(log_files, "Can't make md5sum from", dirname, buf + 2, "", "");
		  fprintf(log_files_accerr, "%s/%s\n", dirname, buf + 2);
		  fflush(log_files_accerr);
		  printf("X");
		  fflush(stdout);
		  s_ferr++;
		} else {
		  dirname[dirnamelen] = 0;
		  emessage(log_files, "new file", dirname, "", md5sum_file, "");
		  fputs(dirname, log_files_new);
		  fputs("\n", log_files_new);
		  fflush(log_files_new);
		  printf("n");
		  fflush(stdout);
		  s_fnew++;
		  // an locale db anfgen, an globale db anfgen
		  if (mode != 2) {
		    localMD5DBchanged = 1;
		    addmd5sum(".md5sums", buf + 2, md5sum_file);
		    if (local == 0) {
		      strcpy(dirname + dirnamelen, buf + 1);
		      strcpy(pname + pnamelen, ".md5sums");
		      addmd5sum(pname, dirname, md5sum_file);
		    }
		  }
		}
	      } else {
		strcpy(md5sum_file, makemd5sum(buf));
		if (md5sum_file[0] == 0) {
		  dirname[dirnamelen] = 0;
		  emessage(log_files, "Can't make md5sum from", dirname, buf + 2, "", "");
		  fprintf(log_files_accerr, "%s/%s\n", dirname, buf + 2);
		  fflush(log_files_accerr);
		  printf("X");
		  fflush(stdout);
		  s_ferr++;
		} else {
		  if (strcmp(md5sum_file, md5sum_db) != 0) {
		    dirname[dirnamelen] = 0;
		    emessage(log_files, "file changed - found via timedate", dirname, buf + 2, md5sum_db, md5sum_file);
		    fprintf(log_files_changed, "%s/%s\n", dirname, buf + 2);
		    fflush(log_files_changed);
		    printf("c");
		    fflush(stdout);
		    s_fchanged++;
		    if (mode != 2) {
		      localMD5DBchanged = 1;
		      // in localer und globeler db ndern
		      changemd5sum(".md5sums", buf + 2, md5sum_file, 0);
		      if (local == 0) {
			strcpy(dirname + dirnamelen, buf + 1);
			strcpy(pname + pnamelen, ".md5sums");
			changemd5sum(pname, dirname, md5sum_file, 1);
		      }
		    }
		  } else {
		    printf(".");
		    fflush(stdout);
		  }
		}
	      }
	    }
	  }
	  pclose(p);
	}
	printf("\n");
	fflush(stdout);
      }
    }

  }
  remove("./.md5sums.timedate");

  // Falls nderungen an einer der beiden lokalen DB vorgenommen,
  // Prfsumme erzeugen und in Haupt DB ablegen
  if (local == 0) {
    if ((mode != 2) && (localMD5DBchanged != 0)) {
      strcpy(pname + pnamelen, ".2md5sums");
      strcpy(dirname + dirnamelen, "/.md5sums");
      strcpy(md5sum_file, makemd5sum(dirname));
      changemd5sum(pname, dirname, md5sum_file, 1);
    }
    if ((mode != 2) && (localMD5FSTATSchanged != 0)) {
      strcpy(dirname + dirnamelen, "/.md5mgr.fstats");
      strcpy(pname + pnamelen, ".2md5mgr.fstats");
      strcpy(md5sum_file, makemd5sum(dirname));
      changemd5sum(pname, dirname, md5sum_file, 1);
    }
  }
}


// hilfe ausgeben, exit
helpexit() {
  fprintf(stderr, "find <dir> -type d ... | md5mgr [relax] [simple] [trustnew] <mode> <project>\n"
	  "  <dir>             : Path to the observe area\n"
	  "  ...               : extra find-options to exclude directorys\n"
	  "  <project>:path    : Path of the Projectfiles (Main Databases)\n"
	  "            local   : Don't use Main Databases. Only local DBs\n"
	  "  <mode>: full      : testing all files with md5sum and change MD5DBs\n"
	  "          test      : full test without changing MD5DBs\n"
	  "          simple add: test only new files, touched files and removed files\n"
	  "          add       : finds also files via: size, timestamp, uid, gid and mode\n"
	  "  trustnew          : use DBs in new dirs. Normall new DBs will be removed.\n"
	  "                    : This option can used to recreate a main DB.\n"
	  "  relax             : Use this after copy trust dirs. It will ignore:\n"
	  "                        change time (not modification time), gid, uid, mode\n");
  exit(1);
}


main(argc, argv)
  int argc;
  char *argv[];
{
  int i;
  int BUFSIZE = MAXLENFILENAME + MAXLELDIRNAME + 50;
  // Namens-Buffer fr aktuelles Verzeichnis und Dateien in diesem
  char buf[BUFSIZE];
  // Namens-Buffer fr Haupt DBs und Logfiles
  char pnamebuf[MAXLENFILENAME];
  int pnamelen;
  int mode = -1; // full, add or test
  int local = 0;
  int simple = 0;
  int poffset = 0;
  int trustnew = 0;
  int relax = 0;
  char *PWD = (char *) get_current_dir_name();

  s_dirs = s_files = s_ferr = s_fnew = s_fchanged = s_fmove = s_frem = s_filesi = s_fileso = s_mcha = 0;

  puts(blacky_ver_info);
  printf("\n"

"This program is free software; you can redistribute it and/or modify it under\n"
"the terms of the GNU General Public License as published by the Free Software\n"
"Foundation; either version 2 of the License, or (at your option) any later\n"
"version.\n"
"This program is distributed in the hope that it will be useful, but WITHOUT\n"
"ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS\n"
"FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.\n\n");

  pnamebuf[0] = 0;

  // Parameter auswerten
  for (i = 1; i < argc; i++) {
    if (strcmp(argv[i], "simple") == 0) {
      simple = 1;
      continue;
    }
    if (strcmp(argv[i], "trustnew") == 0) {
      trustnew = 1;
      continue;
    }
    if (strcmp(argv[i], "full") == 0) {
      mode = 0;
      continue;
    }
    if (strcmp(argv[i], "add") == 0) {
      mode = 1;
      continue;
    }
    if (strcmp(argv[i], "test") == 0) {
      mode = 2;
      continue;
    }
    if (strcmp(argv[i], "relax") == 0) {
      relax = 1;
      continue;
    }

    if (pnamebuf[0] == 0) {
      if (strcmp(argv[i], "local") == 0) {
	local = 1;
	
	pnamelen = strlen(PWD) + 6;
	strcpy(pnamebuf, PWD);
	strcat(pnamebuf, "/local");
	continue;
      }

      if (argv[i][0] == '/') {
	strcpy(pnamebuf, argv[i]);
	continue;
      }
    }

    fprintf(stderr, "Unknown parameter: %s\n(project-path must absolut.)\n\n", argv[i]);
    helpexit();
  }


  if (mode < 0) {
    fprintf(stderr, "Mode not suported!\n");
    helpexit();
  }

  pnamelen = strlen(pnamebuf);
  if (pnamebuf[0] != '/') {
    printf("project path must be absolute!\n");
    fflush(stdout);
    exit(1);
  }

  if ((local == 1) && (mode != 2)) {
    printf("*** WARNING: Local should only uesd with test-mode because other\n" \
	   "             mode can cause differences between local and main db.\n"
	   "             You can recreate the main db with trastnew. Remove the\n"
	   "             old db before!\n");
    printf("(waiting 3 seconds)...\n");
    fflush(stdout);
    sleep(3);
  }

  // make sure that there are no old (not synced) cache file (sync them)
  // transfer cached date to db
  strcpy(pnamebuf + pnamelen, ".md5sums");
  changemd5sum_transcache(pnamebuf);

  strcpy(pnamebuf + pnamelen, ".2md5sums");
  changemd5sum_transcache(pnamebuf);

  strcpy(pnamebuf + pnamelen, ".2md5mgr.fstats");
  changemd5sum_transcache(pnamebuf);


  logs_init(pnamebuf, pnamelen, argc, argv);

  // Durcharbeiten aller Verzeichnisse
  while (!feof(stdin)) {
    buf[0] = 0;
    fgets(buf, BUFSIZE - 1, stdin);
    killenter(buf);
    if (buf[0] != 0) {
      dodir(pnamebuf, pnamelen, buf, mode, local, simple, trustnew, relax);
      chdir(PWD);
    }
  }


  // transfer cached date to db
  strcpy(pnamebuf + pnamelen, ".md5sums");
  changemd5sum_transcache(pnamebuf);

  strcpy(pnamebuf + pnamelen, ".2md5sums");
  changemd5sum_transcache(pnamebuf);

  strcpy(pnamebuf + pnamelen, ".2md5mgr.fstats");
  changemd5sum_transcache(pnamebuf);


  if (local == 1) {
    printf("\nLocal mode can't notice the follow things:\n  - removed directorys\n  - manipulations in the md5sum files\n\n");
  } else {
    // berprfen, ob alle Verzeichnisse, welche in der DB eingetragen
    // sind, auch vorhanden sind
  reopen_db2:
    {
      FILE *md5sum_dir, *p;
      char md5sum_file[33], md5sum_db[33];
      
      strcpy(pnamebuf + pnamelen, ".2md5sums");
      if ((md5sum_dir = fopen(pnamebuf, "r")) == NULL) {
	//      perror(pnamebuf);
	//      exit(1);
      } else {
	
	while (!feof(md5sum_dir)) {
	  buf[0] = 0;
	  fgets(buf, BUFSIZE - 1, md5sum_dir);
	  killenter(buf);
	
	  if ((p = fopen(buf + 34, "r")) == NULL) {
	    int len;
	    // in db lschen
	    strcpy(pnamebuf + pnamelen, ".2md5sums");
	    if (mode != 2) {
	      changemd5sum(pnamebuf, buf + 34, "", 0);
	    }
	    len = strlen(buf);
	    if (strcmp(buf + len - 9, "/.md5sums") == 0) {
	      buf[len - 9] = 0;
	    } else {
	    }
	    //	    (strstr(buf, "/md5sums"))[0] = 0;
	    emessage(log_files, "directory removed", buf + 34, "", "", "");
	    printf("directory has been removed: %s\n", buf + 34);
	    strcpy(pnamebuf + pnamelen, ".2md5mgr.fstats");
	    if (mode != 2) {
	      changemd5sum(pnamebuf, buf + 34, "", 0);
	    }
	    strcpy(pnamebuf + pnamelen, ".md5sums");
	    if (mode != 2) {
	      removing_all_entrys_of_a_directory_from_main_db(pnamebuf, pnamelen, pnamebuf, buf + 34);
	      fclose(md5sum_dir);
	      goto reopen_db2;
	    }
	  } else {
	    fclose(p);
	  }
	}
	fclose(md5sum_dir);
      }
    }
  }

  // Statistik ausgeben:
  printf("dirs scanned:      %06d\n", s_dirs);
  printf("files scanned:     %06d (without removed-files)\n", s_files);
  printf("(-) files ignored: %06d\n", s_filesi);
  printf("(.) files OK:      %06d\n\n", s_fileso);

  printf("md5sums canged/new:%06d\n", s_mcha);
  printf("(X) read error:    %06d\n", s_ferr);
  printf("(n)ew files:       %06d\n", s_fnew);
  printf("(c)hanged files:   %06d\n", s_fchanged);
  printf("(r)emoved:         %06d\n", s_frem);
  //  printf("(m)oved:           %06d (possible moved, same md5sum, also listen under removed)\n\n", s_fmove);

  pnamebuf[pnamelen] = 0;
  printf("See %s.log.* for more details!\n", pnamebuf);

  logs_down();
}
